import java.io.*;
import java.util.ArrayList;

// An InstanceSet is a set of instances to be used in a machine
// learning problem, generally as either a training set or a test set.
class InstanceSet {
    // All instances in the InstanceSet share the same set of
    // attributes, stored here as an AttributeSet
    AttributeSet attributeSet = new AttributeSet();

    // An ArrayList containing all instances in the InstanceSet
    ArrayList<Instance> instances = new ArrayList<Instance>();
    
    String commentStart = "%";
    
    // Construct an InstanceSet by reading a .arff file with the given filename
    public InstanceSet(String inputFilename) throws FileNotFoundException,IOException,Exception
    {
        parseInputFile(inputFilename);
    }

    private void parseInputFile(String inputFilename) throws FileNotFoundException,IOException,Exception
    {
        BufferedReader reader = new BufferedReader(new FileReader(inputFilename));

        // Input file comprises a "preamble" followed by "data".
        // Data segment is demarcated by a line consisting of the string "@data".
        
        // We make everything case insensitive by transforming to
        // lowercase before doing anything else.
        
        // We parse the preamble first, constructing attributes and
        // the like.
        String line = reader.readLine().toLowerCase();
        line = line.toLowerCase();
        while(!parsePreambleLine(line)) {
            line = reader.readLine();
            line = line.toLowerCase();
        }
        attributeSet.setDefaultClassAttribute();

        System.out.println("\nFinished processing preamble, attribute set is:");
        this.attributeSet.print();

        // Now parse the data
        line = reader.readLine().toLowerCase();
        while(line != null) {
            line = line.toLowerCase();
            parseDataLine(line);
            line = reader.readLine();
        }
        reader.close();

        System.out.println("\nFinished processing data, " + instances.size() + 
                           " instances found:");
        for( Instance instance: instances )
            instance.print();
        System.out.println();
    }

    // Return true if this line is the start of the data segment,
    // otherwise return false
    private boolean parsePreambleLine(String line) throws Exception
    {
        if(line.equals(""))
            return false;
        else if(line.startsWith(this.commentStart))
            return false;
        else if(line.startsWith("@relation"))
            return false;
        else if(line.startsWith("@attribute")) {
            parseAttribute(line);
            return false;
        }
        else if(line.startsWith("@data"))
            return true;
        else
            throw new Exception("unexpected line: " + line);
    }

    private void parseAttribute(String line)
    {
        // split on whitespace to extract attribute name
        String[] split_line = line.split("\\s+");
        String attribute_name = split_line[1];

        // Extract the part in braces, which is the list of possible
        // attribute values.
        // To keep things simple, we will split on all brace
        // characters, thus assuming that the only place either
        // brace character appears is at the start and end of the
        // attribute value list
        split_line = line.split("[{}]");
        String value_list = split_line[1];

        // To get the actual values out of value_list, split on commas
        // and whitespace
        String[] values = value_list.split("[,\\s]+");

        // Construct the attribute and add it to the attribute set
        Attribute attribute = new Attribute(attribute_name, values);
        this.attributeSet.addAttribute(attribute);
    }

    private void parseDataLine(String line)
    {
        // ignore comments
        if(line.startsWith(this.commentStart))
            return;

        // Line of data values should be comma-separated, so split on
        // commas and whitespace
        String[] values = line.split("[,\\s]+");
        Instance instance = new Instance(values);
        this.instances.add(instance);
    }
}